import logging
from datetime import timedelta, datetime
from typing import List, Tuple, Union
from django.conf import settings
from ipware.ip import get_client_ip  # type: ignore  # pytype: disable=import-error
from j2fa.errors import TwoFactorAuthError
from j2fa.forms import TwoFactorForm
from j2fa.models import TwoFactorSession
from j2fa.helpers import j2fa_make_code, j2fa_phone_filter, j2fa_send_sms
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.http import HttpRequest
from django.shortcuts import render, redirect
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext as _
from django.views.generic import TemplateView
from django.core.mail import send_mail

logger = logging.getLogger(__name__)


class TwoFactorAuth(TemplateView):
    template_name = "j2fa/ask-code2.html"
    logout_view_name = "admin:logout"
    default_next_view_name = "admin:index"
    max_failed_attempts = 5
    max_failed_attempts_hours = 24

    def get_user_email(self, user: User) -> str:
        """
        Allow user-specific customization of email address to receive the 2FA code.
        Return empty string if user is not supposed to get codes via email but email sending is enabled (J2FA_SEND_TO_EMAIL=True).
        :param user: User
        :return: Email address (if email should be used to send the code, empty otherwise)
        """
        return user.email

    def get_user_phone(self, user: User) -> str:
        """
        Returns User's phone number. By default uses User.profile.phone.
        :param user: User
        :return: str
        """
        if user.is_authenticated and hasattr(user, "profile") and hasattr(user.profile, "phone"):  # type: ignore
            return user.profile.phone  # type: ignore
        return ""

    def make_2fa_code(self) -> str:
        """
        Makes 2FA code.
        By default makes 4-6 digit integer code as str.
        :return: str
        """
        return j2fa_make_code()

    def get_context_data(self, **kw):
        request = self.request
        assert isinstance(request, HttpRequest)

        next_url = request.POST.get("next") if request.POST else None
        if not next_url:
            next_url = request.GET.get("next")
        if not next_url:
            next_url = request.META.get("HTTP_REFERER")
        if not next_url:
            next_url = reverse(self.default_next_view_name)

        cx = {
            "form": TwoFactorForm(data=request.POST or None),
            "next": next_url,
        }
        for k, v in kw.items():
            if v:
                cx[k] = v
        return cx

    def get(self, request, *args, **kwargs):
        user = request.user
        if not user.is_authenticated:
            return redirect(self.logout_view_name)

        cx = self.get_context_data()
        try:
            ses = self.get_session(request)
            channel = request.GET.get("channel") or ""
            next_path = request.GET.get("next") or ""
            if channel:
                self.send_code(ses, channel=channel)
                return redirect(reverse("j2fa-obtain-auth") + f"?msg-resent={channel}&next={next_path}")
            msg_resent = request.GET.get("msg-resent") or ""
            if msg_resent:
                cx["info"] = _("msg.resent")
                cx["alt_channel"] = "I" if msg_resent == "M" else "M"
        except ValidationError as e:
            cx["error"] = " ".join(e.messages)

        return render(request, self.template_name, cx)

    def count_failed_attempts(self, user, since: datetime) -> int:
        return TwoFactorSession.objects.all().filter(user=user, created__gt=since, archived=False).count()

    def send_sms_code(self, phone: str, msg: str, channel: str = "") -> int:
        """
        Send SMS code.
        Args:
            phone: str
            msg: str
            channel: str

        Returns:
            HTTP response status, >300 on error.
        """
        res = j2fa_send_sms(phone, msg, channel=channel)
        return res.status_code

    def send_email_code(
        self, subject: str, body: str, sender: Union[str, Tuple[str, str]], recipient_list: List[Union[str, Tuple[str, str]]], fail_silently: bool = False
    ):
        """
        Emails the OTP code
        Args:
            subject:
            body:
            sender:
            recipient_list:
            fail_silently:

        Returns:
            None
        """
        send_mail(subject, body, sender, recipient_list, fail_silently=fail_silently)

    def send_code(self, ses: TwoFactorSession, channel: str = ""):
        logger.info("2FA: %s -> '%s' (%s) %s", ses.code, ses.phone, ses.user, channel)
        send_by_email = hasattr(settings, "J2FA_SEND_TO_EMAIL") and settings.J2FA_SEND_TO_EMAIL and ses.email
        if settings.SMS_TOKEN and ses.phone:
            status_code = self.send_sms_code(ses.phone, ses.code, channel=channel)
            if status_code >= 300 and hasattr(settings, "EMAIL_HOST") and settings.EMAIL_HOST:
                logger.warning("SMS sending failed to %s (%s), trying to send code by email", ses.phone, ses.user)
                send_by_email = settings.J2FA_FALLBACK_TO_EMAIL if hasattr(settings, "J2FA_FALLBACK_TO_EMAIL") else False
        if send_by_email:
            logger.info("2FA (email): %s -> %s (%s)", ses.code, ses.email, ses.user)
            subject = settings.SMS_SENDER_NAME + ": " + _("One time login code")
            body = _("j2fa.code.email.body").format(code=ses.code)
            sender = settings.DEFAULT_FROM_EMAIL
            recipient = ses.email
            self.send_email_code(subject, body, sender, [recipient], fail_silently=False)

    def get_session(self, request: HttpRequest, force: bool = False) -> TwoFactorSession:
        user, ip, user_agent, phone, email = self.get_session_const(request)
        ses_id = request.session.get("j2fa_session")
        ses = TwoFactorSession.objects.filter(id=ses_id).first() if ses_id else None
        assert ses is None or isinstance(ses, TwoFactorSession)
        if not ses or not ses.is_valid(user, ip, user_agent) or force:
            since = now() - timedelta(hours=self.max_failed_attempts_hours)
            if self.count_failed_attempts(user, since) > self.max_failed_attempts:
                raise TwoFactorAuthError(_("too.many.failed.attempts"))

            ses = TwoFactorSession.objects.create(
                user=user,
                ip=ip,
                user_agent=user_agent[:512],
                phone=phone,
                email=email,
                code=self.make_2fa_code(),
            )
            assert isinstance(ses, TwoFactorSession)
            self.send_code(ses)
            request.session["j2fa_session"] = ses.id
        return ses

    def get_session_const(self, request: HttpRequest):
        user = request.user
        ip = get_client_ip(request)[0]
        if ip is None and settings.DEBUG:
            ip = "127.0.0.1"
        user_agent = request.META["HTTP_USER_AGENT"]
        phone = j2fa_phone_filter(self.get_user_phone(user))  # type: ignore
        email = self.get_user_email(user)  # type: ignore
        if not phone:
            raise TwoFactorAuthError(_("your.phone.number.missing.from.system"))
        return user, ip, user_agent, phone, email

    def post(self, request, *args, **kwargs):  # pylint: disable=unused-argument
        cx = self.get_context_data()
        form = cx["form"]
        assert isinstance(form, TwoFactorForm)

        if form.is_valid():
            try:
                ses = self.get_session(request)
                assert isinstance(ses, TwoFactorSession)
                code = form.cleaned_data["code"]
                user, ip, user_agent, phone, email = self.get_session_const(request)
                logger.info("2FA: Post %s %s %s %s %s vs %s", user, ip, user_agent, phone, email, ses.code)
                if ses.code != code:
                    self.get_session(request, force=True)
                    raise TwoFactorAuthError(_("Invalid code, sending a new one."))

                logger.info("2FA: Pass %s / %s", user, ses)
                ses.activate()

                return redirect(cx.get("next"))
            except TwoFactorAuthError as exc:
                form.add_error(None, exc)
            except Exception as exc:
                form.add_error(None, exc)

        return render(request, self.template_name, cx)



# Generated by Django 4.0.4 on 2022-11-29 21:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("j2fa", "0003_auto_20180807_2202"),
    ]

    operations = [
        migrations.AddField(
            model_name="twofactorsession",
            name="email",
            field=models.EmailField(blank=True, default="", max_length=254),
        ),
    ]



import logging
from j2fa.models import TwoFactorSession
from django.conf import settings
from ipware.ip import get_client_ip  # type: ignore  # pytype: disable=import-error
from django.http import HttpRequest

logger = logging.getLogger(__name__)


def auth_2fa(request: HttpRequest, user, reason: str = ""):
    """
    By-pass 2FA requirement for this user session.
    Call auth.login() before this function if the user is not logged in yet.
    :param request:
    :param user:
    :param reason: Optional reason for by-passing 2FA, e.g. "bank ID". Just stored in log.
    :return:
    """
    ip = get_client_ip(request)[0]
    if ip is None and settings.DEBUG:
        ip = "127.0.0.1"
    user_agent = request.META["HTTP_USER_AGENT"]
    ses = TwoFactorSession.objects.create(
        user=user,
        ip=ip,
        user_agent=user_agent[:512],
        phone="",
        email=user.email,
        code="",
    )
    assert isinstance(ses, TwoFactorSession)
    request.session["j2fa_session"] = ses.id
    ses.activate()
    logger.info("User %s (IP %s) 2FA requirement by-passed (reason: %s)", user, ip, reason)



import logging
import re
from random import randint, choice
import requests
from django.conf import settings


J2FA_PHONE_FILTER = re.compile(r"[^+0-9]")

logger = logging.getLogger(__name__)


def j2fa_make_code(min_digits: int = 4, max_digits: int = 6):
    charset = "123456789"
    code = ""
    for i in range(randint(min_digits, max_digits)):  # pylint: disable=unused-variable
        code += choice(charset)
    return code


def j2fa_phone_filter(v: str) -> str:
    return J2FA_PHONE_FILTER.sub("", v)


def j2fa_send_sms(phone: str, message: str, sender: str = "", channel: str = "", **kwargs):
    """Sends SMS via Kajala Group SMS API. Contact info@kajala.com for access.

    Args:
        phone: Phone number
        message: Message to be esnd
        sender: Sender (max 11 characters)
        **kw: Variable key-value pairs to be sent to SMS API

    Returns:
        Response from requests.post
    """
    if not hasattr(settings, "SMS_TOKEN") or not settings.SMS_TOKEN:
        raise Exception("Invalid configuration: settings.SMS_TOKEN missing")
    if not sender:
        sender = settings.SMS_SENDER_NAME
    if not sender:
        raise Exception("Invalid configuration: settings.SMS_SENDER_NAME missing")
    headers = {
        "Content-Type": "application/json",
        "Authorization": "Token " + settings.SMS_TOKEN,
    }
    data = {
        "dst": j2fa_phone_filter(phone),
        "msg": message,
        "src": sender,
    }
    if channel:
        data["channel"] = channel
    for k, v in kwargs.items():
        data[k] = v
    res = requests.post("https://sms.kajala.com/api/sms/", json=data, headers=headers)
    logger.info("HTTP POST https://sms.kajala.com/api/sms/ status %s", res.status_code)
    return res



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-11-02 01:58
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="TwoFactorSession",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(blank=True, db_index=True, default=django.utils.timezone.now)),
                ("user_agent", models.CharField(max_length=512)),
                ("ip", models.GenericIPAddressField(db_index=True)),
                ("phone", models.CharField(db_index=True, max_length=32)),
                ("code", models.CharField(max_length=8)),
                ("active", models.BooleanField(db_index=True, default=False)),
                ("archived", models.BooleanField(db_index=True, default=False)),
                (
                    "user",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name="+", to=settings.AUTH_USER_MODEL),
                ),
            ],
        ),
    ]



import logging
from ipware.ip import get_client_ip  # type: ignore  # pytype: disable=import-error
from j2fa.models import TwoFactorSession
from django.contrib.auth.models import User
from django.contrib.sessions.backends.base import SessionBase
from django.http import HttpRequest
from django.conf import settings
from django.shortcuts import redirect
from django.urls import resolve, reverse

logger = logging.getLogger(__name__)


class Ensure2FactorAuthenticatedMiddleware:
    """
    Ensures that User is 2-factor authenticated.
    Place after session init.
    By default, uses User.profile.require_2fa boolean to check if 2FA is required for the user.
    """

    excluded_routes = [
        "j2fa-obtain-auth",
        "logout",
    ]
    require_2fa_view = "j2fa-obtain-auth"

    def __init__(self, get_response=None):
        self.get_response = get_response

    def is_2fa_required(self, user: User) -> bool:
        """
        Checks 2FA should be used for this User.
        By default uses user.profile.require_2fa boolean.
        You can override this if custom toggle location is needed.
        Defaults to True.
        :param user: User
        :return: bool
        """
        if hasattr(settings, "J2FA_ENABLED") and not settings.J2FA_ENABLED:  # type: ignore
            return False
        if hasattr(user, "is_hijacked") and user.is_hijacked:  # type: ignore
            return False
        if user.is_authenticated and hasattr(user, "profile") and hasattr(user.profile, "require_2fa"):  # type: ignore
            return user.profile.require_2fa  # type: ignore
        return True

    def is_2fa_route(self, path_info: str) -> bool:
        """
        Checks if request path should be require 2FA.
        Default checks route name against excluded_routes (overridable by settings.J2FA_EXCLUDED_ROUTES).
        :return: bool
        """
        route_name = resolve(path_info).url_name
        return route_name not in self.excluded_routes

    def __call__(self, request: HttpRequest):
        user = request.user
        if user.is_authenticated:
            session = request.session
            assert isinstance(session, SessionBase)
            user_agent = request.META.get("HTTP_USER_AGENT") or ""
            ip = get_client_ip(request)[0] or ""
            path_info = request.path_info
            logger.debug("2FA: IP=%s, path_info=%s, route_name=%s", ip, path_info, resolve(path_info).url_name)

            if self.is_2fa_required(user) and self.is_2fa_route(path_info):
                j2fa_session_id = session.get("j2fa_session") or 0
                j2fa_session = TwoFactorSession.objects.filter(id=j2fa_session_id).first()
                assert j2fa_session is None or isinstance(j2fa_session, TwoFactorSession)
                if j2fa_session is None or not j2fa_session.is_valid(user, ip, user_agent) or not j2fa_session.active:
                    logger.info("2FA: route=%s, j2fa_session_id=%s", self.require_2fa_view, j2fa_session_id)
                    return redirect(reverse(self.require_2fa_view) + "?next=" + request.path)

        return self.get_response(request)



import logging
from django.contrib.auth.models import User
from django.db import models
from django.conf import settings
from django.utils.timezone import now
from j2fa.helpers import j2fa_make_code

logger = logging.getLogger(__name__)


class TwoFactorSession(models.Model):
    created = models.DateTimeField(default=now, db_index=True, blank=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="+", on_delete=models.CASCADE)
    user_agent = models.CharField(max_length=512)
    ip = models.GenericIPAddressField(db_index=True)
    phone = models.CharField(max_length=32, db_index=True)
    email = models.EmailField(blank=True, default="")
    code = models.CharField(max_length=8, default=j2fa_make_code, blank=True)
    active = models.BooleanField(default=False, db_index=True, blank=True)
    archived = models.BooleanField(default=False, db_index=True, blank=True)

    def __str__(self):
        return "[{}]".format(self.id)

    @staticmethod
    def check_ip(ip_a: str, ip_b: str) -> bool:
        """
        Only compares first 2 parts of the IP address to avoid issues with frequently changing IPs.
        :param ip_a:
        :param ip_b:
        :return: bool
        """
        return ip_a.split(".")[:2] == ip_b.split(".")[:2]

    def activate(self):
        self.active = True
        self.archived = False
        self.save()
        TwoFactorSession.objects.all().filter(user=self.user).exclude(id=self.id).update(archived=True, active=False)

    def is_valid(self, user: User, ip: str, user_agent: str) -> bool:
        return self.user == user and self.check_ip(self.ip, ip) and self.user_agent[:512] == user_agent[:512]
